#  -*- coding: utf-8 -*-
# *****************************************************************************
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Markus Zolliker <markus.zolliker@psi.ch>
#
# *****************************************************************************

import time
from math import copysign
from frappy.datatypes import BoolType, FloatRange
from frappy.core import Parameter, BUSY
from frappy.lib import merge_status, clamp
from frappy.errors import RangeError


class HasRamp:
    """software ramp"""
    # make sure it is a drivable
    status = Parameter()
    target = Parameter()
    ramp = Parameter('ramp ratge', FloatRange(0), default=0, readonly=False)
    ramp_used = Parameter('False: infinite ramp', BoolType(), default=False, readonly=False)
    setpoint = Parameter('ramping setpoint', FloatRange())
    maxlag = Parameter('max lag between setpoint and value',
                       FloatRange(0, unit='s'), default=60, readonly=False)
    rampinterval = Parameter('interval for changing the setpoint', FloatRange(0, unit='s'),
                             default=1, readonly=False)
    _ramp_status = None
    _last_time = None

    def checkProperties(self):
        unit = self.parameters['value'].datatype.unit
        self.parameters['setpoint'].setProperty('unit', unit)
        self.writeDict['setpoint'] = self.parameters['setpoint'].default
        super().checkProperties()

    def doPoll(self):
        super().doPoll()  # suppose that this is reading value and status
        self.ramp_step(self.target)

    def write_setpoint(self, value):
        # written only once at startup
        self._last_time = time.time()
        self.setpoint = self.read_value()

    def ramp_step(self, target):
        now = time.time()
        setpoint = self.setpoint
        if self._ramp_status is not None:
            if setpoint == target:
                self._ramp_status = None  # at target
            else:
                sign = copysign(1, target - setpoint)
                delay = now - self._last_time
                dif = (setpoint - self.value) * sign
                ramp_sec = self.ramp / 60.0
                if dif < self.maxlag * ramp_sec:
                    setpoint += delay * sign * ramp_sec
                    if sign * (setpoint - target) >= 0:
                        self.setpoint = target
                        self._ramp_status = None  # at target
                    else:
                        self.setpoint = setpoint
                        self._ramp_status = 'ramping'
                    super().write_target(self.setpoint)
                else:
                    self._ramp_status = 'holding'
        self._last_time = now
        self.read_status()

    def read_status(self):
        status = super().read_status()
        if self._ramp_status is None:
            if self.pollInfo.fast_flag:
                self.setFastPoll(False)
            return status
        if self.pollInfo.interval != self.rampinterval:
            self.setFastPoll(True, self.rampinterval)
        return merge_status((BUSY, self._ramp_status), status)

    def write_ramp(self, ramp):
        if ramp:
            self.write_ramp_used(True)
        else:
            raise RangeError('ramp must not 0, use ramp_used = False to disable ramping')
        return ramp

    def write_ramp_used(self, used):
        if used != self.ramp_used:
            self.ramp_used = used
            if self._ramp_status:
                self.write_target(self.target)

    def read_target(self):
        if not self._ramp_status:
            return super().read_target()
        return self.target

    def write_target(self, target):
        if not self.ramp_used:
            self._ramp_status = None
            self.setpoint = target
            super().write_target(target)
            return target
        self._ramp_status = 'changed target'
        v = self.read_value()
        maxdif = self.maxlag * self.ramp / 60
        # setpoint must not differ too much from value
        self.setpoint = clamp(v - maxdif, self.setpoint, v + maxdif)
        self.setFastPoll(True, self.rampinterval)
        self.ramp_step(target)
        return target
